// Copyright (c) FIRST and other WPILib contributors.
// Open Source Software; you can modify and/or share it under the terms of
// the WPILib BSD license file in the root directory of this project.

#include "UsbCameraListener.h"

using namespace cs;

class UsbCameraListener::Impl {};

UsbCameraListener::UsbCameraListener(wpi::Logger& logger, Notifier& notifier) {}

UsbCameraListener::~UsbCameraListener() = default;

void UsbCameraListener::Start() {}

void UsbCameraListener::Stop() {}
